'''MIT License'''

'''Copyright (c) 2022 David A. Smith'''

'''Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:'''

'''The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.'''

'''THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

'''This code converts the output data from Mumax3, originally in .omf file
format, into spacetime plots of each component of the magnetization. It also is
used for calculating the gradient of the phase angle to make plots of the 
superfluid velocity versus time for a single cell (a point along x)'''
 
'''Note: The raw output of Mumax3 is .ovf, and needs to be converted to .omf using
mumax3-convert.'''

'''Author: David A. Smith'''
'''Date Created: 3/26/2020'''

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from math import atan2, sqrt # Used to import quadrant sensitive arctan function, syntax atan2(y,x)
import glob
import os
import cv2
from scipy.fft import fft
from scipy.signal import find_peaks

#------------------------------------------------------------------------------
    
'''Function to create a video using the images of the magnetization'''
def image_to_avi():
    
    img_array = [] # Initialize empty array to store images
    for filename in glob.glob('m0*.jpg'):
        img = cv2.imread(filename) # Read filename
        height, width, layers = img.shape
        size = (width,height) # Determine image size
        img_array.append(img) # Add image to array
        
    # Create the video with the specified filename, fourcc (?), fps rate, and size
    video_name = r'0_SLST_dynamics.avi'
    out = cv2.VideoWriter(video_name,cv2.VideoWriter_fourcc(*'DIVX'),30,size)
        
    # Release video
    for i in range(len(img_array)):
        out.write(img_array[i])
    out.release()

    return

#------------------------------------------------------------------------------

'''Determine the precessional frequency of the magnetization via a fast Fourier
transform. Create a plot of  FFT vs frequency and a file of FFT frequency vs
position along the centerline of the strip.'''
def frequency(Nx,Ny,snapshotrate,numsamples,x,Mx):
    
    # Determine precessional frequency of the magnetization through a FFT

    mx_fft = np.zeros((Nx,Ny),dtype=float)
    freq = np.linspace(0.0,1/(2*snapshotrate*1e-12),numsamples//2)
    for i in range(0,Nx):
        for j in range(0,Ny):
            temp_fft = fft(Mx[i,j,0,:])
            max_value = np.max(np.abs(temp_fft))
            for k in range(0,len(temp_fft)):
                temp_value = np.abs(temp_fft[k])
                if temp_value == max_value:
                    index = k
                    break
                
            frequency = freq[index]
            mx_fft[i,j] = frequency*1e-9 # Hz => GHz
            
    plt.plot(x[:],mx_fft[:,Ny//2])
    plt.xlabel('Position (Cell Number)')
    plt.ylabel('Precessional Frequency (GHz)')
    plt.title('Precessional Frequency of m$_x$ along Centerline = %5.2f GHz' %(frequency*1e-9))
    plt.savefig('0_mx_frequency_vs_position.png')
    plt.close()
    
    f = open('0_mx_frequency_vs_positon.txt','w+')
    f.write('X Position (Cell Number)' + '\t' + 'Frequency (GHz)' + '\n')
    for i in range(0,len(x)):
        f.write('%f\t%f\n' %(x[i],mx_fft[i,Ny//2]))
    f.close()  
    
    return 
    
#------------------------------------------------------------------------------


if __name__ == "__main__":
    
    '''Make sure these items are correct to the simulation before running:
        1) Path directory - parent folder
        2) Runtime and snapshotrate
        3) Nx; Ny; Nz
    '''
    
    # Below values should match those in the MX3 code
    runtime = 40 # Total run time of the simulation in nanoseconds
    snapshotrate = 20 # Interval of snapshots of m, in picoseconds/snapshot
    numsamples = 1 + (1000*runtime//snapshotrate) # Total number of snapshot of m
    Nx = 4000; Ny = 40; Nz = 2 # Should correspond to the same number as SetGridSize()
    Lenx = 10000; Leny = 100; Lenz = 4 # Size of the sample, in nm
    cellx = Lenx/Nx; celly = Leny/Ny; cellz = Lenz/Nz # Cell size, in nm
    
    
    # Specify all files within the desired directory, store as strings
    parent_folder = r'D:\mumax3.10'#'slst-dw\dynamics-vs-length\SL_2um'
    
    # If analyzing multiple simulations use below
    # filelist = [r'\FeV_SAF_2um_with_Tri_Notches_Jc_2.50e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_3.0e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_3.50e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_4.0e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_4.50e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_5e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_6e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_7e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_8e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_9e11',
    #             r'\FeV_SAF_2um_with_Tri_Notches_Jc_10e11']
    
    # # Specify all values of current density used
    # Jc_list = [2.5e11,3.0e11,3.5e11,4.0e11,4.5e11,5e11,6e11,7e11,8e11,9e11,10e11]
    
    # If analyzing one simulation, use below
    filelist = [r'\FeV_SAF_10um_Jc_10e11_40ns']
    
    Jc_list = [10e11]


    # Initialize empty arrays to store data for superfluid velocities
    before = []; during = []; after = []
    before_err = []; during_err = []; after_err = []
    
    mz_before = []; mz_during = []; mz_after = []
    mz_before_err = []; mz_during_err = []; mz_after_err = []
    
    Mz_list = []; Mz_err_list = []
    
    holding_list = [] # Empty list to store peak positions in Mx

    Jc_index = 0
    for name in filelist:
        # Change to directory where the data is stored
        location =  parent_folder + name
        os.chdir(location)
        print(os.getcwd()) # Print current working directory
            
        # Initialize empty arrays to store components of magnetization   
        Mx_array = np.zeros([Nx,Ny,Nz,numsamples],dtype=float)
        My_array = np.zeros([Nx,Ny,Nz,numsamples],dtype=float)
        Mz_array = np.zeros([Nx,Ny,Nz,numsamples],dtype=float)
        
        # Initialize empty array to store displacement values, counting number of cells from left to right
        # as well as time values, counting number of samples and knowing the snapshot rate
        x_array = np.arange(0,Nx,1); y_array = np.arange(0,Ny,1);
        z_array = np.arange(0,Nz,1); t_array = np.arange(0,numsamples,1);
        filename_array = [] # Array containing all files to be analyzed
        l = 0  # Index for counting number of samples, i.e. time
        for filename in glob.glob('*.omf'):
            filename_array.append(filename)
            # Read in raw data and temporarily store as a Pandas DF
            df = pd.read_csv(filename,sep=' ',skiprows=29,skipfooter=2,
                              names=['Mx','My','Mz'],index_col=False,
                              engine='python')
            m_array = df.values # Turns above DF into a numpy array
            
            # Loop through m_array and store values in corresponding arrays in the correct order
            # Correct order: Start at -z, -y, go -x to +x, and then -y to +y , and then -z to +z
            for k in range(0,Nz):
                for j in range(0,Ny):
                    for i in range(0,Nx):
                        Mx_array[i,j,k,l] = m_array[i+j*Nx+k*Nx*Ny,0]
                        My_array[i,j,k,l] = m_array[i+j*Nx+k*Nx*Ny,1]
                        Mz_array[i,j,k,l] = m_array[i+j*Nx+k*Nx*Ny,2]        
            l += 1 # Increment index counting numsamples as time
            if (l%(numsamples//10)) == 0:
                print('Percent complete = %2.0f' %(l/(numsamples/100)))
                        
        # Calculate the in-plane phase angle at each point
        print('Calculating angles...')
        angle_array = np.zeros([Nx,Ny,Nz,numsamples],dtype=float)
        for l in range(0,numsamples):
            for k in range(0,Nz):
                for j in range(0,Ny):
                    for i in range(0,Nx):
                        angle_array[i,j,k,l] = atan2(My_array[i,j,k,l],Mx_array[i,j,k,l])                                      

        # Determine location of peaks in Mx before (100 nm < x < 900 nm) and after (1100 nm < 1900 nm)
        Mx_peaks_prenotch, _ = find_peaks(Mx_array[40:360,Ny//2,0,numsamples//2],height=(0.8,None))
        Mx_peaks_postnotch, _ = find_peaks(Mx_array[440:760,Ny//2,0,numsamples//2],height=(0.8,None))
        wavelengths_prenotch = []; wavelengths_postnotch = []
        for i in range(0,len(Mx_peaks_prenotch)-1):
            wavelength = Mx_peaks_prenotch[i+1] - Mx_peaks_prenotch[i]
            wavelengths_prenotch.append(wavelength)
        for i in range(0,len(Mx_peaks_postnotch)-1):
            wavelength = Mx_peaks_postnotch[i+1] - Mx_peaks_postnotch[i]
            wavelengths_postnotch.append(wavelength)
                
        Jc = Jc_list[Jc_index]
        line = []
        line.append(Jc)
        for i in range(0,len(Mx_peaks_prenotch)):
            line.append(40+Mx_peaks_prenotch[i])
        for i in range(0,len(Mx_peaks_postnotch)):
            line.append(440+Mx_peaks_postnotch[i])
        print(line)
        holding_list.append(line)
        Jc_index += 1

    # Problem: Each line needs to be the same length. But how to best do this when I don't know how many peaks will be in each file?
    # original_holding_list = holding_list
    # num_peaks = len(holding_list[-1])
    # for i in range(0,len(holding_list)):
    #     while len(holding_list[i]) < num_peaks:
    #         holding_list[i].append(0)
            
    # os.chdir(parent_folder)
    # holding_array = np.array(holding_list)
    # header = 'Jc \t Mx Peaks Prenotch \t Mx Peaks Postnotch \t Ignore Zeros'
    # np.savetxt('000_mx_peak_positions_t=10ns.txt',holding_array,fmt='%5.0f',delimiter='\t',header=header)
    
    # print('Complete!')

#------------------------------------------------------------------------------
                        
        # Call function to make a video using output images
        print('Making dynamics video...')
        image_to_avi()
    
# ------------------------------------------------------------------------------

        # Call function to determine precessional frequency    
        print('Calculating precessional frequency...')
        frequency(Nx,Ny,snapshotrate,numsamples,x_array,Mx_array)
        
#------------------------------------------------------------------------------


        print('Making plots and files...')
        
        cell_before = Nx//4; cell_during = Nx//2; cell_after = 3*Nx//4
        
        # Determine superfluid velocity and plot vs. position and time in seperate plots
        if Nz == 1: # Case for single layer ferromagnets
            # Calculate gradient; determine average and standard deviation
            gradient_array = np.gradient(angle_array[:,:,0,:],x_array,y_array,t_array)
            gradx_phi = gradient_array[0]
            '''Hard code fix for spikes in gradient, fix with a better way later'''
            for i in range(0,len(x_array)):
                for j in range(0,len(y_array)):
                    for k in range(0,len(z_array)):
                        for l in range(0,len(t_array)):
                            if gradx_phi[i,j,l] < -2:
                                gradx_phi[i,j,l] += np.pi
            avg_velocity = np.average(gradx_phi[100,Ny//2,(runtime//5):])
            err = np.std(gradx_phi[100,Ny//2,(runtime//5):])
        
            # Plot superfluid velocity vs time for fixed position
            plt.scatter(t_array*snapshotrate*1e-12,gradx_phi[100,Ny//2,:])
            plt.xlim(0,1.0*runtime*1e-9)
            plt.xlabel("Time (s)")
            plt.ylabel('Superfluid Velocity')
            plt.title('Superfluid Velocity vs. Simulation Time\n Avg Velocity = %7.5f $\pm$ %7.5f' %(avg_velocity,err))
            plt.savefig('0_superfluid_velocity_vs_time.png')
            plt.close()
            
            # Plot superfluid velocity vs position for fixed time
            plt.scatter(x_array,gradx_phi[:,Ny//2,numsamples*3//4])
            plt.xlim(0,len(x_array))
            plt.xlabel('X Position (Cell Number)')
            plt.ylabel('Superfluid Velcoity')
            plt.title('Superfluid velocity vs. Position')
            plt.savefig('0_superfluid_velocity_vs_position.png')
            plt.close()
            
            # Average superfluid velocity once in steady state and plot vs position
            temp_average = np.zeros(Nx,dtype=float)
            temp_err = np.zeros(Nx,dtype=float)
            for i in range(0,Nx):
                temp_average[i] = np.average(gradx_phi[i,Ny//2,(numsamples//5):])
                temp_err[i] = np.std(gradx_phi[i,Ny//2,(numsamples//5):])
            plt.scatter(x_array,temp_average)
            plt.xlim(0,len(x_array))
            plt.xlabel('X Position (Cell Number)')
            plt.ylabel('Superfluid Velocity')
            plt.title('Averaged Superfluid Velocity vs. Position')
            plt.savefig('0_superfluid_velocity_averaged_vs_position.png')
            plt.close()            
            
            # Determine superfluid velocity before, during, and after the notched 
            # after reaching a steady state (~2ns, could change with simulation parameters)
            velocity_before = np.average(gradx_phi[cell_before,Ny//2,(runtime//5):])
            velocity_before_err = np.std(gradx_phi[cell_before,Ny//2,(runtime//5):])
            before_mz = np.average(Mz_array[cell_before,Ny//2,0,(runtime//5):])
            before_mz_err = np.std(Mz_array[cell_before,Ny//2,0,(runtime//5):])                
            
            velocity_during = np.average(gradx_phi[cell_during,Ny//2,(runtime//5):])
            velocity_during_err = np.std(gradx_phi[cell_during,Ny//2,(runtime//5):])
            during_mz = np.average(Mz_array[cell_during,Ny//2,0,(runtime//5):])
            during_mz_err = np.std(Mz_array[cell_during,Ny//2,0,(runtime//5):])
            
            velocity_after = np.average(gradx_phi[cell_after,Ny//2,(runtime//5):])
            velocity_after_err = np.std(gradx_phi[cell_after,Ny//2,(runtime//5):])
            after_mz = np.average(Mz_array[cell_after,Ny//2,0,(runtime//5):])
            after_mz_err = np.std(Mz_array[cell_after,Ny//2,0,(runtime//5):])
            
            # Plot the various superfluid velocities vs simulation time and save figure
            plt.plot(t_array*snapshotrate*1e-12,gradx_phi[cell_before,Ny//2,:],label='Before')
            plt.plot(t_array*snapshotrate*1e-12,gradx_phi[cell_during,Ny//2,:],label='During')
            plt.plot(t_array*snapshotrate*1e-12,gradx_phi[cell_after,Ny//2,:],label='After')
            plt.xlabel('Time (s)')
            plt.ylabel('Superfluid Velocity')
            plt.title('Superfluid Velocity at Various Points vs. Time \n'
                          'Before = %7.4f $\pm$ %7.4f \t'
                          'During = %7.4f $\pm$ %7.4f \n'
                          'After = %7.4f $\pm$ %7.4f'
                          %(velocity_before,velocity_before_err,velocity_during,
                            velocity_during_err,velocity_after,velocity_after_err))
            plt.xlim(0,1.0*runtime*1e-9)
            plt.legend(loc='best')
            plt.savefig('0_superfluid_velocity_vs_time_comparison.png')
            plt.close()
            
            # Write the data to a file
            f = open('0_superfluid_velocity_vs_time_comparison.txt','w+')
            f.write('Time' + '\t' + 'Velocity Before' + '\t' + 'Velocity During'
                    + '\t' + 'Velocity After' + '\n')
            f.write('ns' + '\t' + 'rad/cell' + '\t' + 'rad/cell' + '\t' + 'rad/cell' + '\n')
            for i in range(0,len(t_array)):
                f.write('%f\t%7.4f\t%7.4f\t%7.4f\n' %(t_array[i]/100,
                                                  gradx_phi[cell_before,Ny//2,i],
                                                  gradx_phi[cell_during,Ny//2,i],
                                                  gradx_phi[cell_after,Ny//2,i])
                )
            f.close()
                
            # Add velocity data to lists to be written to a file later
            before.append(velocity_before); before_err.append(velocity_before_err)
            during.append(velocity_during); during_err.append(velocity_during_err)
            after.append(velocity_after); after_err.append(velocity_after_err)
            mz_before.append(before_mz); mz_before_err.append(before_mz_err)
            mz_during.append(during_mz); mz_during_err.append(during_mz_err)
            mz_after.append(after_mz); mz_after_err.append(after_mz_err)
            
            # Write data to files
            f = open('0_superfluid_velocity_vs_time.txt','w+') # Filename
            f.write('Time' + '\t' + 'Superfluid Velocity' + '\n') # Column titles
            f.write('ns' + '\t' + 'rad/cell' + '\n') # Units
            for i in range(0,len(t_array)):
                f.write('%f\t%f\n' %(t_array[i]/100,gradx_phi[100,Ny//2,i])) # Data
            f.close()

            f = open('0_superfluid_velocity_vs_position.txt','w+') # Filename
            f.write('Position' + '\t' + 'Superfluid Velocity' + '\n') # Column titles
            f.write('Cell Number' + '\t' + 'rad/cell' + '\n')
            for i in range(0,len(x_array)):
                f.write('%f\t%f\n' %(x_array[i],gradx_phi[i,Ny//2,numsamples*3//4]))
            f.close()   
        
            f = open('0_superfluid_velocity_averaged_vs_position.txt','w+') 
            f.write('Position' + '\t' + 'Averaged Velocity' + '\t' + 'err' + '\n')
            f.write('Cell Number' + '\t' + 'rad/cell' + '\t' + 'rad/cell' + '\n')
            for i in range(0,len(x_array)):
                f.write('%f\t%f\t%f\n' %(x_array[i],temp_average[i],temp_err[i]))
            f.close()
            
            print('Calculating DW position...')
            dw_pos_list = []
            max_num_peaks = 0
            min_height = 0.8
        
            # Calculate gradient of mx along x axis
            grad_mx = np.gradient(Mx_array[:,:,0,:],x_array,y_array,t_array)
            gradx_mx = grad_mx[0]
            
            for t in range(0,numsamples):
                        
                # Find peaks in the gradient of mx; correspond to DW positions
                peaks, _ = find_peaks(np.abs(My_array[:,Ny//2,0,t]),height=(min_height,None))
                # height = (minimum height of peak, maximum height of peak)
                
                # Get number of peaks (i.e. DWs) and store the maximum number 
                if len(peaks) > max_num_peaks:
                    max_num_peaks = len(peaks)
                            
            # Create an empty array to store the positions of every DW at a time t        
            all_DW_pos = np.zeros((numsamples,max_num_peaks),dtype=float)
            for t in range(0,numsamples):
                
                # Find all the peaks again, but this time savea all the positions at each time            
                peaks, _ = find_peaks(np.abs(My_array[:,Ny//2,0,t]),height=(min_height,None))
                
                # t = time, i = peak number
                for i in range(0,peaks.size):
                    all_DW_pos[t,i] = peaks[i]      
            
            # Save the array with all DW positions as a text file 
            description = 'Each row is one time interval; each column is a DW position'
            np.savetxt('all_DW_positions_vs_time.txt', all_DW_pos, fmt = '%5.0f',
                        delimiter = ' ', header = description)
            
        
        else: # Case for Nz = 2, primarily designed for synthetic antiferromagnets
        
            # Calculate gradient; determine average and standard deviation
            gradient_array = np.gradient(angle_array,x_array,y_array,z_array,t_array)
            gradx_phi = gradient_array[0]
            '''Hard code fix for spikes in gradient, fix with a better way later'''
            for i in range(0,len(x_array)):
                for j in range(0,len(y_array)):
                    for k in range(0,len(z_array)):
                        for l in range(0,len(t_array)):
                            if gradx_phi[i,j,k,l] < -2:
                                gradx_phi[i,j,k,l] += np.pi
            avg_velocity = np.average(gradx_phi[100,Ny//2,0,(runtime//5):])
            err = np.std(gradx_phi[100,Ny//2,0,(runtime//5):])
        
            # Plot superfluid velocity vs time for fixed position
            plt.scatter(t_array*snapshotrate*1e-12,gradx_phi[100,Ny//2,0,:])
            plt.xlim(0,1.0*runtime*1.0e-9)
            plt.xlabel("Time (s)")
            plt.ylabel('Superfluid Velocity')
            plt.title('Superfluid Velocity vs. Simulation Time\n Avg Velocity = %7.5f $\pm$ %7.5f rad/s' %(avg_velocity,err))
            plt.savefig('0_superfluid_velocity_vs_time.png')
            plt.close()
            
            # Plot superfluid velocity vs position for fixed time
            plt.scatter(x_array,gradx_phi[:,Ny//2,0,numsamples*3//4])
            plt.xlim(0,len(x_array))
            plt.xlabel('X Position (Cell Number)')
            plt.ylabel('Superfluid Velocity')
            plt.title('Superfluid Velocity vs. Position')
            plt.savefig('0_superfluid_velocity_vs_position.png')
            plt.close()  
    
            # Average superfluid velocity once in steady state and plot vs position
            temp_average = np.zeros(Nx,dtype=float)
            temp_err = np.zeros(Nx,dtype=float)
            for i in range(0,Nx):
                temp_average[i] = np.average(gradx_phi[i,Ny//2,0,(numsamples//5):])
                temp_err[i] = np.std(gradx_phi[i,Ny//2,0,(numsamples//5):])
            plt.scatter(x_array,temp_average)
            plt.xlim(0,len(x_array))
            plt.xlabel('X Position (Cell Number)')
            plt.ylabel('Superfluid Velocity')
            plt.title('Averaged Superfluid Velocity vs. Position')
            plt.savefig('0_superfluid_velocity_averaged_vs_position.png')
            plt.close()              

            # Determine superfluid velocity before, during, and after the notched 
            # after reaching a steady state (changes with total simulation runtime)
            velocity_before = np.average(gradx_phi[cell_before,Ny//2,0,(runtime//5):])
            velocity_before_err = np.std(gradx_phi[cell_before,Ny//2,0,(runtime//5):])
            before_mz = np.average(Mz_array[cell_before,Ny//2,0,(runtime//5):])
            before_mz_err = np.std(Mz_array[cell_before,Ny//2,0,(runtime//5):])
            
            velocity_during = np.average(gradx_phi[cell_during,Ny//2,0,(runtime//5):])
            velocity_during_err = np.std(gradx_phi[cell_during,Ny//2,0,(runtime//5):])
            during_mz = np.average(Mz_array[cell_during,Ny//2,0,(runtime//5):])
            during_mz_err = np.std(Mz_array[cell_during,Ny//2,0,(runtime//5):])
            
            velocity_after = np.average(gradx_phi[cell_after,Ny//2,0,(runtime//5):])
            velocity_after_err = np.std(gradx_phi[cell_after,Ny//2,0,(runtime//5):])
            after_mz = np.average(Mz_array[cell_after,Ny//2,0,(runtime//5):])
            after_mz_err = np.std(Mz_array[cell_after,Ny//2,0,(runtime//5):])
            
            # Plot the various superfluid velocities vs simulation time and save figure
            plt.plot(t_array*snapshotrate*1e-12,gradx_phi[cell_before,Ny//2,0,:],label='Before')
            plt.plot(t_array*snapshotrate*1e-12,gradx_phi[cell_during,Ny//2,0,:],label='During')
            plt.plot(t_array*snapshotrate*1e-12,gradx_phi[cell_after,Ny//2,0,:],label='After')
            plt.xlabel('Time (s)')
            plt.ylabel('Superfluid Velocity')
            plt.title('Superfluid Velocity at Various Points vs. Time \n'
                      'Before = %7.4f $\pm$ %7.4f \t'
                      'During = %7.4f $\pm$ %7.4f \n'
                      'After = %7.4f $\pm$ %7.4f'
                      %(velocity_before,velocity_before_err,velocity_during,
                        velocity_during_err,velocity_after,velocity_after_err))
            plt.xlim(0,1.0*runtime*1e-9)
            plt.legend(loc='best')
            plt.savefig('0_superfluid_velocity_vs_time_comparison.png')
            plt.close()
        
            # Write the data to a file
            f = open('0_superfluid_velocity_vs_time_comparison.txt','w+')
            f.write('Time' + '\t' + 'Velocity Before' + '\t' + 'Velocity During'
                    + '\t' + 'Velocity After' + '\n')
            f.write('ns' + '\t' + 'rad/cell' + '\t' + 'rad/cell' + '\t' + 'rad/cell' + '\n')
            for i in range(0,len(t_array)):
                f.write('%f\t%7.4f\t%7.4f\t%7.4f\n' %(t_array[i]/100,
                                                  gradx_phi[cell_before,Ny//2,0,i],
                                                  gradx_phi[cell_during,Ny//2,0,i],
                                                  gradx_phi[cell_after,Ny//2,0,i])
                )
            f.close()
            
            # Add velocity data to lists to be written to a file later
            before.append(velocity_before); before_err.append(velocity_before_err)
            during.append(velocity_during); during_err.append(velocity_during_err)
            after.append(velocity_after); after_err.append(velocity_after_err)
            mz_before.append(before_mz); mz_before_err.append(before_mz_err)
            mz_during.append(during_mz); mz_during_err.append(during_mz_err)
            mz_after.append(after_mz); mz_after_err.append(after_mz_err)  

            # Write data to files
            f = open('0_superfluid_velocity_vs_time.txt','w+') # Filename
            f.write('Time' + '\t' + 'Superfluid Velocity' + '\n') # Column titles
            f.write('ns' + '\t' + 'rad/cell' + '\n') # Units
            for i in range(0,len(t_array)):
                f.write('%f\t%f\n' %(t_array[i]/100,gradx_phi[100,Ny//2,0,i])) # Data
            f.close()
        
            f = open('0_superfluid_velocity_vs_position.txt','w+') # Filename
            f.write('Position' + '\t' + 'Superfluid Velocity' + '\n') # Column titles
            f.write('Cell Number' + '\t' + 'rad/cell' + '\n')
            for i in range(0,len(x_array)):
                f.write('%f\t%f\n' %(x_array[i],gradx_phi[i,Ny//2,0,numsamples*3//4]))
            f.close()     
            
            f = open('0_superfluid_velocity_averaged_vs_position.txt','w+') 
            f.write('Position' + '\t' + 'Averaged Velocity' + '\t' + 'err' + '\n')
            f.write('Cell Number' + '\t' + 'rad/cell' + '\t' + 'rad/cell' + '\n')
            for i in range(0,len(x_array)):
                f.write('%f\t%f\t%f\n' %(x_array[i],temp_average[i],temp_err[i]))
            f.close()   

            # Plot Mx for one point in the sample, both top and bottom layers, vs. time
            plt.plot(t_array*snapshotrate*1e-12,Mx_array[Nx//4,Ny//2,0,:],label='Bottom Layer')
            plt.plot(t_array*snapshotrate*1e-12,Mx_array[Nx//4,Ny//2,1,:],label='Top Layer')
            plt.xlim(0,1.0*runtime*1.0e-9)
            plt.ylim(-1.25,1.25)
            plt.xlabel('Time (s)')
            plt.ylabel('M$_x$')
            plt.legend(loc='best')
            plt.title('M$_x$ for Top and Bottom Layers vs. Simulation Time')
            plt.savefig('0_mx_vs_time_comparison_top_bottom_layers.png')
            plt.close()     
            
            print('Calculating DW position...')
            dw_pos_list = [] # Create empty list to store DW positions
            min_height = 0.8
            
            max_num_peaks = 0 # Initialize maximum number of peaks
            
            # Calculate gradient of mx along x axis
            grad_mx = np.gradient(Mx_array,x_array,y_array,z_array,t_array)
            gradx_mx = grad_mx[0]
            
            for t in range(0,numsamples):
                
                # Find peaks in the gradient of mx; correspond to DW positions
                peaks, _ = find_peaks(np.abs(My_array[:,Ny//2,1,t]),
                                      height=(min_height,None))
                
                # Get number of peaks (i.e. DWs) and store the maximum number 
                if len(peaks) > max_num_peaks:
                    max_num_peaks = len(peaks)
            
            # Create an empty array to store the positions of every DW at a time t        
            all_DW_pos = np.zeros((numsamples,max_num_peaks),dtype=float)
            for t in range(0,numsamples):
                
                # Find all the peaks again, but this time save all the positions at each time            
                peaks, _ = find_peaks(np.abs(My_array[:,Ny//2,1,t]),
                                      height=(min_height,None))
                
                # t = time, i = peak number
                for i in range(0,peaks.size):
                    all_DW_pos[t,i] = peaks[i]
                                
            # Save the array with all DW positions as a text file
            description = 'Each row is one time interval; each column is a DW position'
            np.savetxt('0_all_DW_positions_vs_time.txt', all_DW_pos, fmt = '%5.0f',
                delimiter = ' ', header = description)
            
#------------------------------------------------------------------------------
        
        # Plot Mx for two points in the sample on the bottom layer vs. time
        plt.plot(t_array*snapshotrate*1e-12,Mx_array[Nx//4,Ny//2,0,:],label='Cell %i' %(Nx//4))
        plt.plot(t_array*snapshotrate*1e-12,Mx_array[3*Nx//4,Ny//2,0,:],label='Cell %i' %(3*Nx//4))
        plt.xlim(0,1.0*runtime*1.0e-9)
        plt.ylim(-1.5,1.5)
        plt.xlabel('Time (s)')
        plt.ylabel('M$_x$')
        plt.title('M$_x$ for Select Cells vs. Simulation Time')
        plt.legend(loc='best')
        plt.savefig('0_mx_vs_time_two_cells_bottom_layer.png')
        plt.close()
       
#------------------------------------------------------------------------------
            
        # Plot Mz vs position for fixed time
        plt.scatter(x_array,Mz_array[:,Ny//2,0,numsamples*3//4])
        plt.xlim(0,len(x_array))
        plt.xlabel('X Position (Cell Number)')
        plt.ylabel('M$_z$')
        plt.title('M$_z$ vs Position')
        plt.savefig('0_mz_vs_position.png')
        plt.close()

#------------------------------------------------------------------------------
    
        # Average Mz once in steady state and plot versus position.
        # Average in detection region and append to list
        Mz_average = np.zeros(Nx,dtype=float)
        Mz_average_err = np.zeros(Nx,dtype=float)
        for i in range(0,Nx):
            Mz_average[i] = np.average(Mz_array[i,Ny//2,0,(numsamples//5):])
            Mz_average_err[i] = np.std(Mz_array[i,Ny//2,0,(numsamples//5):])
        plt.scatter(x_array,Mz_average)
        plt.xlim(0,Nx)
        plt.ylim(0,1.0)
        plt.xlabel('X Position (Cell Number)')
        plt.ylabel('m$_z$')
        plt.title('Averaged m$_z$ vs Position')
        plt.savefig('0_mz_averaged_vs_position.png')
        plt.close()
        Mz_detector = np.average(Mz_average[(Nx-40):Nx]) # Hardcoded size of detector region in cell size
        Mz_detector_err = np.std(Mz_average[(Nx-40):Nx]) 
        Mz_list.append(Mz_detector)
        Mz_err_list.append(Mz_detector_err)
        
#------------------------------------------------------------------------------
            
        # Print magnetization data for a cell to a file within the directory
        f = open('0_regional_m_vs_time.txt','w+') # Filename
        f.write('Time' + '\t' + 'Mx' + '\t' + 'My' + '\t' + 'Mz' + '\n') # Column titles
        f.write('ps' + '\t' + ' ' + '\t' + ' ' + '\t' + ' ' + '\n') # Units
        for i in range(0,len(t_array)): # Data
            f.write('%f\t%f\t%f\t%f\n' %(t_array[i]*snapshotrate,
                                          Mx_array[Nx//4,Ny//2,0,i],
                                          My_array[Nx//4,Ny//2,0,i],
                                          Mz_array[Nx//4,Ny//2,0,i])
                                          )
        f.close()

#------------------------------------------------------------------------------
        
        # Print averaged mz vs position to a file
        f = open('0_mz_averaged_vs_position.txt.','w+') # Filename
        f.write('Position' + '\t' + 'Mz' + '\t' + 'err' + '\n') # Column titles
        f.write('Cell Number' + '\t' + ' ' + '\t' + ' ' + '\n')
        for i in range(0,len(x_array)): # Data
            f.write('%f\t%f\t%f\n' %(x_array[i],Mz_average[i],Mz_average_err[i]))
        f.close()
        
#------------------------------------------------------------------------------
        
        # Print magnetization data for a time to a file within the directory
        f = open('0_regional_m_vs_position.txt','w+') # Filename
        f.write('Position' + '\t' + 'Mx' + '\t' + 'My' + '\t' + 'Mz' + '\n') # Column titles
        f.write('Cell Number' + '\t' + ' ' + '\t' + ' ' + '\t' + ' ' + '\n') # Units
        for i in range(0,len(x_array)): # Data
            f.write('%f\t%f\t%f\t%f\n' %(x_array[i],
                                          Mx_array[i,Ny//2,0,numsamples*3//4],
                                          My_array[i,Ny//2,0,numsamples*3//4],
                                          Mz_array[i,Ny//2,0,numsamples*3//4])
                                          )
        f.close()
        
#------------------------------------------------------------------------------
        '''
        1) Find peaks in the absolute value of the My_array at a given time t.
        2) If location of a peak is past a certain x position (e.g. x = 1800 nm) or within a range of positions
            at that time t, save that time t. This will be the time of arrival for a specific DW.
        3) Repeat this for all times
        4) Plot DW time of arrival as a series of delta functions vs. simulation time. The location
            of each delta function should correspond to the arrival time of DWs.
        5) Take the Fourier transform of the train of delta functions.
        '''
    
        print('Doing FFT related items for DW TOA...')
        # Sampling rate, snapshotrate = 10ps
        # Sampling frequency, fs = 1/snapshotrate = 100 GHz
        # Nyquist frequency, fN = fs/2 = 50 GHz
        
        # Compute FFT of my near the end of the strip. Plot and write to file
        steady_start = 401 # Designate the start of the steady state, in snapshots (time in ns = snapshotrate/100)
        freq = np.linspace(0.0,1/(2*snapshotrate*1e-12),numsamples//2) # Create frequency spectra
        my_fft = fft(np.abs(My_array[9*Nx//10,Ny//2,0,:])) # Compute FFT of my near the end of the strip, in the steady state
        plt.plot(freq*1e-9,np.abs(my_fft[:numsamples//2]),label='Absolute Value') 
        my2_fft = fft(My_array[9*Nx//10,Ny//2,0,:])
        plt.plot(freq*1e-9,np.abs(my2_fft[:numsamples//2]),label='Raw')
        my3_fft = fft((My_array[9*Nx//10,Ny//2,0,:])**2)
        plt.plot(freq*1e-9,np.abs(my3_fft[:numsamples//2]),label='Squared')
        plt.legend(loc='best')
        plt.xlabel('Frequency (GHz)')
        plt.ylabel('FFT of m$_y$')
        plt.title('Frequency of DW Arrivals')
        plt.savefig('0_my_fft.png')
        plt.close()
        
        # Plot My and absolute value of My, compare periods/frequencies
        plt.plot(t_array[steady_start:],np.abs(My_array[9*Nx//10,Ny//2,0,steady_start:]),label='Absolute Value')
        plt.plot(t_array[steady_start:],My_array[9*Nx//10,Ny//2,0,steady_start:],label='Raw')
        plt.plot(t_array[steady_start:],(My_array[9*Nx//10,Ny//2,0,steady_start:])**2,label='Squared')
        plt.xlim(steady_start,numsamples)
        plt.ylim(-1.25,1.25)
        plt.legend(loc='best')
        plt.xlabel('Time (snapshots)')
        plt.ylabel('m$_y$')
        plt.close()
        
        # Write my vs time at the end of the strip to file
        f = open('0_various_my_vs_time.txt','w+')
        f.write('Time (ns)' + '\t' + 'my' + '\t' + 'Abs of my' + '\t' + 'my Squared' + '\n')
        for i in range(0,len(t_array)):
            f.write('%f\t%f\t%f\t%f\n' %(t_array[i]/100,My_array[9*Nx//10,Ny//2,0,i],np.abs(My_array[9*Nx//10,Ny//2,0,i]),(My_array[9*Nx//10,Ny//2,0,i])**2))
        f.close()
        
        
        # Write FFT of my to a file
        f = open('0_my_fft.txt','w+')
        f.write('Frequency (GHz)' + '\t' + 'FFT of my' + '\t' + 'FFT of |my|' + '\t' + 'FFT of my^2' + '\n')
        for i in range(0,len(freq)):
            f.write('%f\t%f\t%f\t%f\n' %(1e-9*freq[i],np.abs(my2_fft[i]),np.abs(my_fft[i]),np.abs(my3_fft[i])))
        f.close()
        
        # Note the time of arrival for DWs.Plot |my| vs time for a fixed position and find all peaks;  
        # these times are when the DWs reach that position. Plot time of arrival as a series of delta function spikes,
        # save plots, and write to file.
        time_of_arrival = np.zeros(numsamples,dtype=float) # Initialize empty array
        min_height = 0.8 # Designate minimum height for the peak search algorithm
        peaks, _ = find_peaks(np.abs(My_array[9*Nx//10,Ny//2,0,:]),height=(min_height,None))       
        for t in range(0,numsamples):
            if t in peaks[:]:
                time_of_arrival[t] = 1
        toa_fft = fft(time_of_arrival[:])
        
        # Plot the time of arrival
        plt.plot(t_array[:],time_of_arrival[:],label='DW TOA')
        plt.legend(loc='best')
        plt.ylim(-0.1,1.5)
        plt.xlabel('Time (ns)')
        plt.ylabel('DW Time of Arrival')
        plt.savefig('0_DW_time_of_arrival.png')
        plt.close()
        
        '''Need to add in freq as a column before FFT of TOA'''
        # Write  time of arrival to a file
        f = open('0_DW_time_of_arrival.txt.','w+')
        f.write('Time (ns)' + '\t' + 'TOA Signal' + '\t' + 'FFT of TOA' + '\n')
        for i in range(0,len(t_array[:])):
            f.write('%f\t%f\t%f\n' %(t_array[i]/100,time_of_arrival[i],np.abs(toa_fft[i])))
        f.close()
        
        # Compute the FFT of the DW time of arrival. Compare with the FFT of my
        plt.plot(freq*1e-9,np.abs(toa_fft[:numsamples//2]),label='FFT of TOA')
        plt.plot(freq*1e-9,np.abs(my_fft[:numsamples//2]),label='FFT of My')
        plt.legend(loc='best')
        plt.xlabel('Frequency (GHz)')
        plt.ylabel('FFT of Time of Arrival')
        plt.savefig('0_DW_time_of_arrival_fft.png')
        plt.close()
    
#------------------------------------------------------------------------------
        
        if len(filelist) > 1: print('Moving on to the next folder...')
        
    if len(filelist) > 1:

        # Use lists of Jc, velocity values before, during, and after notch and
        # write the data to a file.
        print('Creating final set of files...')
        os.chdir(parent_folder)
        f = open('0_superfluid_velocity_comparison.txt','w+')
        f.write('Current Density' + '\t' + 'Before' + '\t' + 'err' + '\t'
                + 'During' + '\t' + 'err' + '\t' + 'After' + '\t' + 'err'
                    + '\n')
        f.write('A/m2' + '\t' + 'rad/cell' + '\t' + 'rad/cell' + '\t'
                + 'rad/cell' + '\t' + 'rad/cell' + '\t' + 'rad/cell' + '\t'
                + 'rad/cell' + '\n')
        for (Jc,b,b_err,d,d_err,a,a_err) in zip(Jc_list,before,before_err,during,during_err,after,after_err):
            f.write('%E\t%f\t%f\t%f\t%f\t%f\t%f\n' %(Jc,b,b_err,d,d_err,a,a_err))
        f.close()
        
        f = open('0_mz_comparison.txt','w+')
        f.write('Current Density' + '\t' + 'Before' + '\t' + 'err' + '\t'
                + 'During' + '\t' + 'err' + '\t' + 'After' + '\t' + 'err'
                + '\n')
        f.write('A/m2' + '\t' + ' ' + '\t' + ' ' + '\t' + ' ' + '\t' + ' '
                + '\t' + ' ' + '\t' + ' '+ '\n')
        for(Jc,b,b_err,d,d_err,a,a_err) in zip(Jc_list,mz_before,mz_before_err,mz_during,mz_during_err,mz_after,mz_after_err):
            f.write('%E\t%f\t%f\t%f\t%f\t%f\t%f\n' %(Jc,b,b_err,d,d_err,a,a_err))
        f.close()
    
        
        # Write averaged Mz in detection region vs. current density to a file
        f = open('0_mz_detector_vs_Jc.txt','w+')
        f.write('Current Density' + '\t' + 'Mz in Detector' + '\t' + 'err' + '\n')
        f.write('A/m2' + '\t' + ' ' + '\t' + ' ' + '\n')
        for (Jc,Mz,err) in zip(Jc_list,Mz_list,Mz_err_list):
            f.write('%E\t%f\t%f\n' %(Jc,Mz,err))
        f.close()
        
        original_holding_list = holding_list
        num_peaks = len(holding_list[-1])
        for i in range(0,len(holding_list)):
            while len(holding_list[i]) < num_peaks:
                holding_list[i].append(0)
                
        os.chdir(parent_folder)
        holding_array = np.array(holding_list)
        header = 'Jc \t Mx Peaks Prenotch \t Mx Peaks Postnotch \t Ignore Zeros'
        np.savetxt('000_mx_peak_positions_t=10ns.txt',holding_array,fmt='%5.0f',delimiter='\t',header=header)
    
    print('Complete!')
        # print('Complete!')
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

    '''This section contains extra bits of code that are designed to be run
    seperate from the rest of the code.'''
  
#------------------------------------------------------------------------------
     
    # # Ensure peak search algorithm is working as expected  
    # peaks, _ = find_peaks(My_array[:,Ny//2,1,500],height=(min_height,None))
    # plt.plot(peaks[-1], My_array[peaks[-1],Ny//2,1,500],'bo',label='First DW in Train')
    # plt.plot(peaks[:-1],My_array[peaks[:-1],Ny//2,1,500],'rx',label='Other DWs')
    # plt.plot(x_array[:],My_array[:,Ny//2,1,500],'k--',label='m$_y$')
    # plt.xlabel('X Position (Cell Number)')
    # plt.ylabel('m$_y$')
    # plt.legend(loc='best')
    # plt.xlim(0,Nx)
    # plt.ylim(-1.5,1.5)
    
#------------------------------------------------------------------------------
    
    # # Create a plot comparing the magnitude of the FFT at different
    # # x positions along the centerline of the nanostrip
    # plt.plot(freq*1e-9,np.abs(fft(Mx_array[Nx//4,Ny//2,0,:)),label='x = 500 nm')
    # plt.plot(freq*1e-9,np.abs(fft(Mx_array[Nx//2,Ny//2,0,:)),label='x = 1000 nm')
    # plt.plot(freq*1e-9,np.abs(fft(Mx_array[3*Nx//4,Ny//2,0,:)),label='x = 1500 nm')
    # plt.xlabel('Frequency (GHz)')
    # plt.ylabel('FFT of m$_x$')
    # plt.xlim(0,50)
    # plt.legend(loc='best')
    # plt.savefig('0_mx_FFT_comparison.png')
    # plt.close()
    
    # # Write to file the FFT of mx at two different x positions
    # f = open('0_mx_FFT_comparison.txt','w+')
    # f.write('Frequency (GHz)' + '\t' + 'FFT1' + '\t' + 'FFT2' + '\t' + 'FFT3' + '\n')
    # fft1 = np.abs(fft(Mx_array[Nx//4,Ny//2,0,:]))
    # fft2 = np.abs(fft(Mx_array[Nx//2,Ny//2,0,:]))
    # fft3 = np.abs(fft(Mx_array[3*Nx//4,Ny//2,0,:]))
    # for i in range(0,len(freq)):
    #     f.write('%f\t%f\t%f\t%f\n' %(1e-9*freq[i],fft1[i],fft2[i],fft3[i]))
    # f.close() 
    
#------------------------------------------------------------------------------
    
    # # Determine the mismatch between DW positions and the length scale on which it decays
    # bottom_peaks = []; top_peaks = []; DW_pos_diff = []
    # for t in range(0,numsamples):
    #     peaks0, _ = find_peaks(np.abs(My_array[:,Ny//2,0,t]),height=(0.8,None))
    #     peaks1, _ = find_peaks(np.abs(My_array[:,Ny//2,1,t]),height=(0.8,None))
    #     if len(peaks0) > 0:
    #         bottom_peaks.append(peaks0[0]*2.5)
    #         top_peaks.append(peaks1[0]*2.5)
    #         DW_pos_diff.append((2.5)*(peaks1[0]-peaks0[0]))
    #     else:
    #         bottom_peaks.append(0)
    #         top_peaks.append(0)
    #         DW_pos_diff.append(0)
            
    # # Create plot of DW position in both the top and bottom layers
    # plt.plot(t_array[:]/100,bottom_peaks[:],label='Bottom Layer')
    # plt.plot(t_array[:]/100,top_peaks[:],label='Top Layer')
    # plt.xlabel('Time (ns)')
    # plt.ylabel('DW Position (nm)')
    # plt.legend(loc='best')
    # plt.savefig('0_DW_position_both_layers.png')
    # plt.close()

    # # Create a subplot that plots the DW position in each layer in an individual subplot
    # # Bottom layer plot
    # plt.subplot(2,1,1)
    # plt.plot(t_array[:]/100,bottom_peaks[:],'b-',label='Bottom Layer')
    # plt.xlabel('Time (ns)')
    # plt.ylabel('DW Position (nm)')
    # plt.legend(loc='best')
    
    # # Top layer plot
    # plt.subplot(2,1,2)
    # plt.plot(t_array[:]/100,top_peaks[:],'r-',label='Top Layer')
    # plt.xlabel('Time (ns)')
    # plt.ylabel('DW Position (nm)')
    # plt.legend(loc='best')
    # plt.savefig('0_DW_position_both_layers_subplots.png')
    # plt.close()
    
    # # Plot and save figure of the difference in DW positions between the two layers
    # plt.plot(t_array[:]/100,DW_pos_diff[:])
    # plt.xlabel('Time (ns)')
    # plt.ylabel('Difference in DW Position (nm)')
    # plt.title('Difference in DW Positions Between Top and Bottom Layer in SAF')
    # plt.savefig('0_DW_position_difference.png')
    # plt.close()
    
#------------------------------------------------------------------------------

    # # Determine DW canting angle (arctan(mz/(mx^2+my^2))) as a function of DW position or simulation time
    
    # # First find the DW position in the top and bottom layers
    # DW_bottom_positions = []; DW_top_positions = [] # Empty lists to store DW positions
    # DW_bottom_cant_angles = []; DW_top_cant_angles = [] # Empty lists to store DW canting angles
    # for t in range(0,numsamples):
    #     bottom_peaks, _ = find_peaks(np.abs(My_array[:,Ny//2,0,t]),height=(0.8,None))
    #     top_peaks, _ = find_peaks(np.abs(My_array[:,Ny//2,1,t]),height=(0.8,None))
    #     if len(bottom_peaks) > 0:
    #         DW_bottom_pos = bottom_peaks[0]
    #         DW_top_pos = top_peaks[0]
    #     else:
    #         DW_bottom_pos = 0
    #         DW_top_pos = 0
    #     DW_bottom_positions.append(DW_bottom_pos*cellx) # Save DW positions, in nm
    #     DW_top_positions.append(DW_top_pos*cellx)
    
    #     # Determine DW canting angle at that position for both top and bottom layers
    #     DW_bottom_angle = atan2(Mz_array[DW_bottom_pos,Ny//2,0,t],(Mx_array[DW_bottom_pos,Ny//2,0,t]**2+My_array[DW_bottom_pos,Ny//2,0,t]**2))
    #     DW_bottom_cant_angles.append(DW_bottom_angle*180/np.pi)
    #     DW_top_angle = atan2(Mz_array[DW_top_pos,Ny//2,1,t],(Mx_array[DW_top_pos,Ny//2,1,t]**2+My_array[DW_top_pos,Ny//2,1,t]**2))
    #     DW_top_cant_angles.append(DW_top_angle*180/np.pi)
    #     # print('DW canting angle = %5.2f deg at t = %4.0f. DW Position = %4.0f nm.' %(DW_angle*180/np.pi,t,DW_pos*2.5))
        
    # # Plot DW canting angle vs. DW position
    # plt.plot(DW_bottom_positions,DW_bottom_cant_angles,label='Bottom Layer')
    # plt.plot(DW_top_positions,DW_top_cant_angles,label='Top Layer')
    # plt.xlabel('DW Position (nm)')
    # plt.ylabel('DW Canting Angle (deg)')
    # plt.title('DW Canting Angle for DWs in Top and Bottom Layer in SAF')
    # plt.legend(loc='best')
    # plt.savefig('0_DW_canting_angle_both_layers.png')
    # plt.close()
    
    # # Write DW canting angle for both layers to a file
    # f = open('0_DW_canting_angle_both_layers.txt','w+')
    # f.write('Bottom DW Pos' + '\t' + 'Bottom DW Angle' + '\t' + 'Top DW Pos' + '\t' + 'Top DW Angle' + '\n')
    # f.write('nm' + '\t' + 'deg' + '\t' + 'nm' + '\t' + 'deg' + '\n')
    # for i in range(0,len(DW_bottom_positions)):
    #     f.write('%f\t%f\t%f\t%f\n' %(DW_bottom_positions[i],DW_bottom_cant_angles[i],DW_top_positions[i],DW_top_cant_angles[i]))
    # f.close()
    
#------------------------------------------------------------------------------

    # # Determine DW width by fitting with a tanh function
    # def tanh_func(x,x0,w):
    #     return -np.tanh((x-x0)/w)
    
    # from scipy.optimize import curve_fit
    
    # DW_width = [] # Initialize an empty list to store DW width vs. simulation time
    
    # p0 = [300,30] # Initial guess values for DW center position and width, in cell number
    # ub = [len(x_array)*cellx,len(x_array)*cellx/2] # Define upper limit
    # lb = [0,0] # Define lower limit
    # t0 = 50 # Define starting point in time for curve_fit, in snapshots

    # # Check that the curve fitting is doing a decent job
    # popt, pcov = curve_fit(tanh_func,x_array,Mx_array[:,Ny//2,1,400],p0=p0,
    #                         bounds = [lb,ub])
    # plt.plot(x_array[:]*cellx,Mx_array[:,Ny//2,1,400],label='raw')
    # plt.plot(x_array[:]*cellx,-np.tanh((cellx*(x_array[:]-popt[0]))/(cellx*popt[1])),label='fit')
    # plt.xlabel('X Position (nm)')
    # plt.ylabel('m$_x$')
    # plt.legend(loc='best')
    # plt.title('Domain Wall Width, $\Delta$ = %4.1f nm' %(popt[1]*cellx))
    # plt.savefig('0_DW_width_vs_position_fitted_curve.png')
    # plt.close()
    
    # # Fit DW width with a tanh function to determine how the DW width changes with simulation time
    # for t in range(t0,numsamples):
    #     popt, pcov = curve_fit(tanh_func,x_array,Mx_array[:,Ny//2,1,t],p0=p0,
    #                         bounds = [lb,ub])
    #     DW_width.append(popt[1]*cellx)    
    
    # # Plot DW width vs. time and save figure
    # plt.plot(t_array[t0:]/100,DW_width[:])
    # plt.xlabel('Time (ns)')
    # plt.ylabel('DW Width (nm)')
    # plt.title('DW Width as a Function of Simulation Time')
    # plt.savefig('0_DW_width_vs_time_from_t=%3.1f_ns.png' %(t0/100))
    # plt.close()

    # # Write DW width vs. simulation time to a file    
    # f = open('0_DW_width_vs_time_from_t=' + str(t0/100) + '_ns.txt','w+')
    # f.write('Time' + '\t' + 'DW Width' + '\n')
    # f.write('ns' + '\t' + 'nm' + '\n')
    # for i in range(t0,len(t_array)):
    #     f.write('%f\t%f\n' %(t_array[i]/100,DW_width[i-t0]))
    # f.close()

#------------------------------------------------------------------------------

    # Attempt to calculate the wavelength at various positions for a dispersion curve
    # Can relate the excitation current density to the precessional frequency, 
    # but how to best determine what the corresponding wavelength is? It's position dependent!    
    

    # Mx_peaks_prenotch, _ = find_peaks(Mx_array[40:360,Ny//2,0,numsamples//2],height=(0.8,None))
    # Mx_peaks_postnotch, _ = find_peaks(Mx_array[440:760,Ny//2,0,numsamples//2],height=(0.8,None))
    # wavelengths_prenotch = []; wavelengths_postnotch = []
    # for i in range(0,len(Mx_peaks_prenotch)-1):
    #     wavelength = Mx_peaks_prenotch[i+1] - Mx_peaks_prenotch[i]
    #     wavelengths_prenotch.append(wavelength)
    # for i in range(0,len(Mx_peaks_postnotch)-1):
    #     wavelength = Mx_peaks_postnotch[i+1] - Mx_peaks_postnotch[i]
    #     wavelengths_postnotch.append(wavelength)
            
    # Jc = 10e11
    # line = []
    # line.append(Jc)
    # for i in range(0,len(Mx_peaks_prenotch)):
    #     line.append(40+Mx_peaks_prenotch[i])
    # for i in range(0,len(Mx_peaks_postnotch)):
    #     line.append(440+Mx_peaks_postnotch[i])
    # print(line)
    # holding_list = []
    # holding_list.append(line)
    # holding_array = np.array(holding_list)
    # header = 'Jc \t Mx Peaks Prenotch \t Mx Peaks Postnotch'
    # np.savetxt('000_holding_array.txt',holding_array,fmt='%4.0f',delimiter='\t',header=header)
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    